/*
  DamagePerFire.cpp - Recognise damage_per_fire.

  Jason Hood, 18 & 19 June, 2011.

  If damage_per_fire is <= 1, it's treated as a fraction and the weapon will be
  damaged based on its /current/ health (e.g. if hit_pts = 400 and
  damage_per_fire = 0.1, then hit_pts will become 360, 324, 291.6, ...); the
  equipment will be destroyed when hit_pts goes less than 1.
  If damage_per_fire is < 2, the fractional part is multiplied by 100 and one is
  added - this is the number of shots that can be fired (e.g. 1.01 can be fired
  twice; 1.30 [and 1.3!] can be fired 31 times).
  Normally, the equipment is turned into debris; use a negative value to have
  it simply disappear.

  v1.01, 18 January, 2014:
  - fix server detection.

  Install:
	Copy DamagePerFire.dll to your EXE directory and add it to the
	[Libraries] sections of dacom.ini & dacomsrv.ini.
	Requires & assumes the official patch.
*/

#include "Server.h"
#include <map>
#include <cmath>

#define NAKED	__declspec( naked )
#define STDCALL __stdcall


// Detect loading of server in single player.
#define ADDR_NEWGAME ((PDWORD)(0x5a8796+1))

// CELauncher::ConsumeFireResources
#define ADDR_FIRE    ((PBYTE)0x6299623)

// Some sort of server update.
#define ADDR_UPDATE  (server + 0x6d014bf+1 - 0x6ce0000)


DWORD dummy;
#define ProtectX( addr, size ) \
  VirtualProtect( addr, size, PAGE_EXECUTE_READWRITE, &dummy )

#define RELOFS( from, to ) \
  *(PDWORD)(from) = (DWORD)(to) - (DWORD)(from) - 4

#define NEWOFS( from, to, prev ) \
  prev = (DWORD)(from) + *(PDWORD)(from) + 4; \
  RELOFS( from, to )

#define JMP( from, to ) \
  *(PBYTE)(from) = 0xE9; \
  RELOFS( (DWORD)(from)+1, to )


typedef std::multimap<UINT, DamageEntry> DamageMap;
typedef DamageMap::const_iterator DamageMapCIter;

DamageMap damage;


void STDCALL store_damage( PBYTE equip, float damage_per_fire, float hit_pts )
{
  DamageEntry de;

  UINT ship = *(UINT*)(equip+4);
  de.hpid = *(USHORT*)(equip+8);
  de.hit_pts = *(float*)(equip+0x30);

  float min = 0;
  float dmg = fabsf( damage_per_fire );
  if (dmg <= 1)
  {
    de.hit_pts -= de.hit_pts * dmg;
    min = 1;
  }
  else if (dmg < 2)
  {
    de.hit_pts -= hit_pts / ((dmg - 1) * 100 + 1);
  }
  else
  {
    de.hit_pts -= dmg;
  }
  if (de.hit_pts < min)
  {
    de.hit_pts = 0;
  }
  de.fate = (de.hit_pts != 0)	  ? DamageEntry::SubObjFate::FATE_NONE :
	    (damage_per_fire < 0) ? DamageEntry::SubObjFate::FATE_DISAPPEAR
				  : DamageEntry::SubObjFate::FATE_DEBRIS;
  damage.insert( std::pair<UINT, DamageEntry>(ship, de) );
}


NAKED
void Fire_Hook()
{
  __asm {
	mov	edx, [edi+12]		// Archetype
	mov	eax, [edx+0x88] 	// damage_per_fire
	test	eax, eax
	jz	done
	push	dword ptr [edx+0x1c]	// hit_pts
	push	eax
	push	edi
	call	store_damage
  done:
	pop	edi
	pop	esi
	ret
  }
}


void STDCALL add_damage( Server* data )
{
  DamageMapCIter iter = damage.lower_bound( data->ship );
  if (iter == damage.end())
    return;

  DamageList dmg;
  DamageMapCIter end = damage.upper_bound( data->ship );
  for (; iter != end; ++iter)
  {
    dmg.add_damage_entry( iter->second.hpid, iter->second.hit_pts,
			  iter->second.fate );
  }
  data->fn188( dmg );
  damage.erase( data->ship );
}


DWORD Update_Org;

NAKED
void Update_Hook()
{
  __asm {
	push	ecx
	call	add_damage
	mov	ecx, esi
	jmp	Update_Org
  }
}


void PatchServer()
{
  PBYTE server = (PBYTE)GetModuleHandle( "server.dll" );

  // Make sure the patch hasn't already been applied, since something like
  // Console keeps server.dll loaded.
  if (server == NULL || *(PDWORD)ADDR_UPDATE != 0xFFFFFC8C)
    return;

  ProtectX( ADDR_UPDATE, 4 );
  NEWOFS( ADDR_UPDATE, Update_Hook, Update_Org );
}


DWORD New_Game_Org;

NAKED
void New_Game_Hook()
{
  __asm {
	push	ecx
	call	PatchServer
	pop	ecx
	jmp	New_Game_Org
  }
}


void Patch()
{
  ProtectX( ADDR_FIRE, 5 );
  JMP( ADDR_FIRE, Fire_Hook );

  if (IsMPServer())
  {
    // For FLServer.
    PatchServer();
  }
  else
  {
    // Server is loaded every time a game is loaded for single player.
    ProtectX( ADDR_NEWGAME, 4 );
    NEWOFS( ADDR_NEWGAME, New_Game_Hook, New_Game_Org );
  }
}


BOOL WINAPI DllMain( HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved )
{
  if (fdwReason == DLL_PROCESS_ATTACH)
    Patch();

  return TRUE;
}
